#region Preset using directives

using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;

using VirtualLab.Programming;
using VirtualLabAPI.Core.BasicFunctions;
using VirtualLabAPI.Core.Common;
using VirtualLabAPI.Core.DataVisualization;
using VirtualLabAPI.Core.FieldRepresentations;
using VirtualLabAPI.Core.Functions;
using VirtualLabAPI.Core.GeometryDescription;
using VirtualLabAPI.Core.LightPath;
using VirtualLabAPI.Core.Materials;
using VirtualLabAPI.Core.Numerics;
using VirtualLabAPI.Core.Numerics.Region2D;
using VirtualLabAPI.Core.OpticalSystems;
using VirtualLabAPI.Core.Propagation;
using VirtualLabAPI.Core.SupportFunctions;

using VirtualLabAPI.Core.Coating;
using VirtualLabAPI.UI;
using VirtualLabAPI.UI.DataVisualization;
using VirtualLabAPI.UI.FieldVisualization;
using VirtualLabAPI.UI.BasicUI;
using VirtualLabAPI.UI.PropertyBrowser;
using VirtualLabAPI.UI.DataVisualization._1DView;
using VirtualLabAPI.UI.DataVisualization.DataArrayDiagrams;
using System.Collections.ObjectModel;

#endregion

namespace OwnCode {
	public class VLModule {
		public void Run() {

			// LightTrans International GmbH
			// This C# module is a sample file accompanying the use case Space Plate System Made by 
			// Multilayer Stack. The purpose  of this module is to generate a plot of the phase of the 
			// transmission coefficient for TE polarization of the layered space plate as a function of 
			// angle of incidence, and to superimpose on it the equivalent phase curve for
			// free-space propagation, with a propagation distance of d_eff, as found by the 
			// fitting process in the reference publication propsed by Orad Reshef et al 
			// [Reshef, O., DelMastro, M.P., Bearne, K.K.M. et al. An optic to replace space and its 
			// application towards ultra-thin imaging systems. Nat Commun 12, 3512 (2021). 
			// https://doi.org/10.1038/s41467-021-23358-8]  


			/***************** THE USER SHOULD ONLY MODIFY THE VALUES OF PARAMETERS IN THIS SECTION:******************/

			double wavelength = 1550e-9; // The wavelength of the radiation. 
			double angleMin = 0; // The lower bound of the range of angles of incidence for the plot, in degrees. 
			double angleMax = 45; // The upper bound of the range of angles of incidence for the plot, in degrees. 
			int nSamplesAngle = 1001; // The number of samples in the specified range of angles of incidence. 

			/*********************************************************************************************************/


			// Calculate wavenumber in vacuum: 
			double wavenumber = MathFunctions.TwoPi / wavelength;

			// Transform the angle variables into radian: 
			angleMin = MathFunctions.DegreeToRadian(angleMin);
			angleMax = MathFunctions.DegreeToRadian(angleMax);

			// Calculate sampling parameters of angle space: 
			double windowSizeAngleSpace = angleMax - angleMin;
			double samplingDistanceAngleSpace = windowSizeAngleSpace / (nSamplesAngle - 1);

			// Prepare materials for coating (layered space plate): 
			// Load the "LightTrans Defined" category of the materials catalog, 
			// without forcing a reload from disk of the catalog: 
			MaterialsCatalog materialsCatalog = MaterialsCatalog.LoadMaterialsCatalogFromDefaultPath(
				CatalogDefinitionType.LightTransDefined,
				false);

			// From the catalog we just loaded, get the "Vacuum" entry for both the material 
			// in front of the coating (space plate) and behind it: 
			StandardMaterial materialInFrontOfSpacePlate = materialsCatalog["Vacuum"] as StandardMaterial;
			StandardMaterial materialBehindSpacePlate = materialsCatalog["Vacuum"] as StandardMaterial;

			// The space plate itself is composed of layers of different thicknesses of two
			// alternating materials. The values of the refractive indices for these two materials
			// for the wavelength in question are provided in the appendix of the reference publication;
			// we use these values here to create the corresponding materials: 
			StandardMaterial materialOddLayers = new StandardMaterial(
				"siliconForSpacePlate",
				new Complex(3.48985, 0.00982674),
				wavelength);
			StandardMaterial materialEvenLayers = new StandardMaterial(
				"silicaForSpacePlate",
				new Complex(new Complex(1.45611, 0)),
				wavelength);

			// Information about the thicknesses of the different layers obtained 
			// through the optimization process described in the reference publication 
			// is also provided in the appendix. We take those values:
			double[] layerThicknesses = new double[] {
				133.86e-9,
				722.324e-9,
				319.406e-9,
				573.293e-9,
				551.083e-9,
				232.074e-9,
				340.955e-9,
				254.686e-9,
				105.252e-9,
				265.555e-9,
				124.592e-9,
				1032.82e-9,
				145.7e-9,
				239.521e-9,
				313.439e-9,
				252.054e-9,
				371.439e-9,
				168.605e-9,
				125.935e-9,
				747.517e-9,
				105.681e-9,
				1629.52e-9,
				318.601e-9,
				334.673e-9,
				766.62e-9};

			// Construct the actual coating (multi-layer space plate):
			StandardCoating spacePlate = new StandardCoating();
			for (int layerIndex = 0; layerIndex < layerThicknesses.Length; layerIndex++) {
				if (layerIndex % 2 == 0) {
					spacePlate.Add(new CoatingLayer(layerThicknesses[layerIndex], materialOddLayers));
				}
				else {
					spacePlate.Add(new CoatingLayer(layerThicknesses[layerIndex], materialEvenLayers));
				}
			}

			// Generate object to calculate the reflection and transmission coefficients of the 
			// space plate: 
			FresnelCoefficientsCalculation fresnelCoefficientsSpacePlate = new FresnelCoefficientsCalculation(
				materialInFrontOfSpacePlate,
				materialBehindSpacePlate,
				spacePlate,
				CoatingOrientation.BackSide);

			// Generate containers for values of phase curves: 
			ComplexFieldDerivative1D phaseValuesSpacePlate = new ComplexFieldDerivative1D(nSamplesAngle, false, PrecisionMode.Double);
			ComplexFieldDerivative1D phaseValuesFreeSpace = new ComplexFieldDerivative1D(nSamplesAngle, false, PrecisionMode.Double);

			// Generate container for amplitude curve of transmission coefficient (the goal is to lose as 
			// little energy as possible):
			ComplexFieldDerivative1D amplitudeSpacePlate = new ComplexFieldDerivative1D(nSamplesAngle, false, PrecisionMode.Double);

			// The reference publication states that the muti-layer space plate they propose achieves
			// a compression factor of 4.9 (with the corresponding free-space propagation distance being
			// equal to the compression factor times the physical thickness of the space plate). We calculate 
			// the equivalent distance for free-space propagation from this data: 
			double compressionFactor = 4.9;
			double effectiveDistance = compressionFactor * spacePlate.GetInitialCoatingThickness();

			// Calculate offset for phase curve of space plate (the curves are adjusted
			// through a constant-phase offset that ensures the value at 0 is 0. 
			double offsetSpacePlate = fresnelCoefficientsSpacePlate.GetFresnelCoefficientsTETM(
				wavelength,
				0,
				Globals.Defaults.SystemTemperatureInDegreesCelsius,
				Globals.Defaults.AirPressureInPascals)[5];
			// Indices of Fresnel coefficients and their physical meaning: 
			// (0) Reflectance TE
			// (1) Transmittance TE
			// (2) Amplitude reflection coefficient TE
			// (3) Phase reflection coefficient TE
			// (4) Amplitude transmission coefficient TE
			// (5) Phase transmission coefficient TE
			// (6) Reflectance TM
			// (7) Transmittance TM
			// (8) Amplitude reflection coefficient TM
			// (9) Phase reflection coefficient TM
			// (10) Amplitude transmission coefficient TM
			// (11) Phase transmission coefficient TM

			// Calculate offset for free-space propagation (same argument):
			double offsetFreeSpace = wavenumber * effectiveDistance;

			// Calculate the values of the phase curves using a loop over the samples 
			// in the angle space:
			for (int angleCounter = 0; angleCounter < nSamplesAngle; angleCounter++) {
				// Calculate incidence angle for current point (value of abscissa): 
				double angleCoordinate = angleMin + angleCounter * samplingDistanceAngleSpace;
				// Calculate phase value of free-space propagation for current point: 
				phaseValuesFreeSpace[angleCounter] =
					wavenumber * Math.Cos(angleCoordinate) * effectiveDistance - offsetFreeSpace;

				// Calculate Fresnel coefficients of space plate for current point: 
				double[] fresnelCoefficientsSpacePlateCurrentPoint = fresnelCoefficientsSpacePlate.GetFresnelCoefficientsTETM(
					wavelength,
					angleCoordinate,
					Globals.Defaults.SystemTemperatureInDegreesCelsius,
					Globals.Defaults.AirPressureInPascals);
				// From that, get phase value of space plate for current point: 
				phaseValuesSpacePlate[angleCounter] =
					fresnelCoefficientsSpacePlateCurrentPoint[5] - offsetSpacePlate;
				// And also amplitude value of transmission coefficient for current point:
				amplitudeSpacePlate[angleCounter] = fresnelCoefficientsSpacePlateCurrentPoint[4];
			}

			// Unwrap phases: 
			phaseValuesSpacePlate = UnwrapPhase1D(phaseValuesSpacePlate);


			// Set up and display phase plot:
			DataArray1D plotPhaseCurves = new DataArray1D(
				new ComplexField1DArray(new ComplexFieldDerivative1D[] {
					phaseValuesFreeSpace,
					phaseValuesSpacePlate,
					}),
				new PhysicalProperty[] {
					PhysicalProperty.AngleRad,
					PhysicalProperty.AngleRad,
					},
				new string[] {
					"Phase free space",
					"Phase space plate",
					},
				samplingDistanceAngleSpace,
				angleMin,
				PhysicalProperty.AngleDeg,
				"Angle of incidence");


			VL_GUI.ShowDocument(plotPhaseCurves, "Space plate vs. free-space propagation");

			// Configure diagram: 
			IFormWithControlForDataArray1D formDA1DPhase = Globals.ActiveDocumentHistoryInternal.GetLastDocumentWindow(0, DocumentFilter.DataArray1D) as IFormWithControlForDataArray1D;

			ControlDataArray1D_MultiGraph controlMultigraphPhase = formDA1DPhase.CurrentlyVisibleControlDataArray1D as ControlDataArray1D_MultiGraph;
			ControlDataArray1DWithTable controlTablePhase = formDA1DPhase.CurrentlyVisibleControlDataArray1D as ControlDataArray1DWithTable;


			VL_GUI.WriteLineToMessagesTab("number of controls " + controlMultigraphPhase.Controls.Count.ToString());
			VL_GUI.WriteLineToMessagesTab(controlMultigraphPhase.Controls[0].Name);
			VL_GUI.WriteLineToMessagesTab(controlMultigraphPhase.Controls[1].Name);

			controlMultigraphPhase.InterpolationMode = ViewInterpolationMode.RealValuedSmoothing;
			controlMultigraphPhase.LegendVisible = true;
			controlMultigraphPhase.SingleGraphMode = false;
			controlMultigraphPhase.LegendAlignment = AlignmentWithNiceName.TopRight;
			controlMultigraphPhase.SymbolScalingFactor = 0;
			controlMultigraphPhase.LineWidth = 2;


			//set the color of plotted lines
			List<PropertiesFor1DGraph> listOfPropertiesLeftPhase = new List<PropertiesFor1DGraph>() {
							new PropertiesFor1DGraph(0, Color.DarkTurquoise, DiagramSymbolShape.FilledTriangleUp),
							new PropertiesFor1DGraph(1, Color.Black, DiagramSymbolShape.OpenCircle)};

			controlMultigraphPhase.SubsetPropertiesLeft = listOfPropertiesLeftPhase.AsReadOnly();

			ControlDataView1D_MultiGraph controlView1DPhase = controlMultigraphPhase.Controls[1] as ControlDataView1D_MultiGraph;
			controlView1DPhase.LogarithmicScaledData = false;
			controlView1DPhase.ColorOfYAxis_Left = Color.Black;
			controlView1DPhase.ColorOfYAxis_Right = Color.Transparent;
			controlView1DPhase.IsAxisLabelUserDefined_yLeft = true;
			controlView1DPhase.UserDefinedAxisLabel_yLeft = "Phase";
			controlView1DPhase.SetBothYRangesAtOnce(
		 		new PhysicalValueRange(new PhysicalValue(-10, PhysicalProperty.AngleRad), new PhysicalValue(0, PhysicalProperty.AngleRad)),
				new PhysicalValueRange(new PhysicalValue(-10, PhysicalProperty.AngleRad), new PhysicalValue(0, PhysicalProperty.AngleRad))
				);


			Globals.DataDisplayInternal.UpdateForms();




			// Set up and display amplitude plot:
			DataArray1D plotAmplitudeCurves = new DataArray1D(
				new ComplexField1DArray(new ComplexFieldDerivative1D[] {
					amplitudeSpacePlate,
					}),
				new PhysicalProperty[] {
					PhysicalProperty.NoUnit,
					},
				new string[] {
					"Amplitude space plate",
					},
				samplingDistanceAngleSpace,
				angleMin,
				PhysicalProperty.AngleDeg,
				"Angle of incidence");


			VL_GUI.ShowDocument(plotAmplitudeCurves, "Space plate vs. free-space propagation");

			// Configure diagram: 
			IFormWithControlForDataArray1D formDA1DAmplitude = Globals.ActiveDocumentHistoryInternal.GetLastDocumentWindow(0, DocumentFilter.DataArray1D) as IFormWithControlForDataArray1D;

			ControlDataArray1D_MultiGraph controlMultigraphAmplitude = formDA1DAmplitude.CurrentlyVisibleControlDataArray1D as ControlDataArray1D_MultiGraph;
			ControlDataArray1DWithTable controlTableAmplitude = formDA1DAmplitude.CurrentlyVisibleControlDataArray1D as ControlDataArray1DWithTable;


			VL_GUI.WriteLineToMessagesTab("number of controls " + controlMultigraphAmplitude.Controls.Count.ToString());
			VL_GUI.WriteLineToMessagesTab(controlMultigraphAmplitude.Controls[0].Name);
			VL_GUI.WriteLineToMessagesTab(controlMultigraphAmplitude.Controls[1].Name);

			controlMultigraphAmplitude.InterpolationMode = ViewInterpolationMode.RealValuedSmoothing;
			controlMultigraphAmplitude.LegendVisible = true;
			controlMultigraphAmplitude.SingleGraphMode = false;
			controlMultigraphAmplitude.LegendAlignment = AlignmentWithNiceName.TopRight;
			controlMultigraphAmplitude.SymbolScalingFactor = 0;
			controlMultigraphAmplitude.LineWidth = 2;


			//set the color of plotted lines
			List<PropertiesFor1DGraph> listOfPropertiesLeftAmplitude = new List<PropertiesFor1DGraph>() {
							new PropertiesFor1DGraph(0, Color.Black, DiagramSymbolShape.FilledTriangleUp),
							};

			controlMultigraphAmplitude.SubsetPropertiesLeft = listOfPropertiesLeftAmplitude.AsReadOnly();


			ControlDataView1D_MultiGraph controlView1DAmplitude = controlMultigraphAmplitude.Controls[1] as ControlDataView1D_MultiGraph;
			controlView1DAmplitude.LogarithmicScaledData = true;
			controlView1DAmplitude.ColorOfYAxis_Left = Color.Black;
			controlView1DAmplitude.ColorOfYAxis_Right = Color.Transparent;
			controlView1DAmplitude.IsAxisLabelUserDefined_yLeft = true;
			controlView1DAmplitude.UserDefinedAxisLabel_yLeft = "Amplitude";
			controlView1DAmplitude.SetBothYRangesAtOnce(
		 		new PhysicalValueRange(new PhysicalValue(8e-5, PhysicalProperty.NoUnit), new PhysicalValue(1, PhysicalProperty.NoUnit)),
				new PhysicalValueRange(new PhysicalValue(8e-5, PhysicalProperty.NoUnit), new PhysicalValue(1, PhysicalProperty.NoUnit))
				);


			Globals.DataDisplayInternal.UpdateForms();

		} // run



		#region Support functions:
		private ComplexFieldDerivative1D UnwrapPhase1D(ComplexFieldDerivative1D wrappedPhase) {
			for (int i = 1; i < wrappedPhase.SamplingPointsOneD; i++) {
				if (wrappedPhase[i].Re - wrappedPhase[i - 1].Re >= Math.PI) {
					for (int i_sub = i; i_sub < wrappedPhase.SamplingPointsOneD; i_sub++) {
						wrappedPhase[i_sub] = wrappedPhase[i_sub] - MathFunctions.TwoPi;
					} // loop over remaining points
				} // if difference between contiguous values is larger than pi
				else if (wrappedPhase[i].Re - wrappedPhase[i - 1].Re <= -Math.PI) {
					for (int i_sub = i; i_sub < wrappedPhase.SamplingPointsOneD; i_sub++) {
						wrappedPhase[i_sub] = wrappedPhase[i_sub] + MathFunctions.TwoPi;
					} // loop over remaining points
				} // if difference smaller than minus pi
			} // loop over sampling points
			return wrappedPhase;
		}
		#endregion

	} // module
} // namespace
